/// Simple optionally-singleton service locator
///
/// Use [GotIt.registerFactory], [GotIt.registerSingleton]
/// or [GotIt.registerLazySingleton] to register whatever you desire.
///
/// Use [GotIt.get] or call the object like a method ([GotIt.call])
/// to get whatever you registered back.
class GotIt {
  /// Creates an unique instance of [GotIt]
  GotIt();

  /// Returns a singleton instance of [GotIt]
  factory GotIt.singleton() => _instance;

  static final _instance = GotIt();

  // DI/SL Containers are always just Maps, this one isn't a singleton however.
  final _types = <Type, _Entry>{};

  /// Register a method that will return a new instance of T
  /// everytime [get] (or [call]) is called.
  void registerFactory<T>(
    T Function() func, {
    bool allowReassignment = false,
  }) =>
      _register<T>(allowReassignment, false, func);

  /// Register an item that will be returned verbatim
  /// everytime [get] (or [call]) is called.
  void registerSingleton<T>(
    T instance, {
    bool allowReassignment = false,
  }) =>
      _register<T>(allowReassignment, false, instance);

  /// Register a method that will create an instance of T on the first call
  /// to [get] (or [call]), and return that instance from then on.
  void registerLazySingleton<T>(
    T Function() func, {
    bool allowReassignment = false,
  }) =>
      _register<T>(allowReassignment, true, func);

  // The backbone of the library
  void _register<T>(bool allowReassignment, bool isLazy, dynamic value) {
    assert(
      _types[T]?.isReassignable ?? true,
      'You cannot reassign to this type',
    );
    _types[T] = _Entry(allowReassignment, isLazy, value);
  }

  /// Same as [get]
  T call<T>() => get<T>();

  /// Grab and return whatever is associated with T
  T get<T>() {
    final entry = _types[T];

    if (entry == null) {
      throw UnregisteredError<T>();
    }

    if (entry.value is T) {
      return entry.value;
    }

    if (entry.isLazy) {
      _types[T] = entry.activate<T>();
      return _types[T].value;
    }

    return _types[T].value();
  }

  /// Delete all registrations, useful for unit tests.
  void reset() => _types.clear();
}

// Contains all entries and represents their behavior at [GotIt.get] time.
class _Entry {
  const _Entry(this.isReassignable, this.isLazy, this.value);

  final bool isReassignable;
  final bool isLazy;
  // This is explicitly dynamic because Dart has no ADT,
  // otherwise, it would be "T Function() | T"
  final dynamic value;

  /// If the current entry is a lazy singleton, instantiate it.
  _Entry activate<T>() {
    if (value is T Function() && isLazy) {
      return _Entry(isReassignable, false, value());
    } else {
      throw StateError(
        'Attempted to activate a non-lazy or non-function value',
      );
    }
  }
}

/// Thrown whenever a call to [GotIt.get] (or [GotIt.call]) fails to resolve T.
class UnregisteredError<T> extends Error {
  @override
  String toString() => "Type $T hasn't been registered.";
}
